"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Decoder = void 0;
var tslib_1 = require("tslib");
var CrdtDecoder_1 = require("../../util/binary/CrdtDecoder");
var clock_1 = require("../../clock");
var PatchBuilder_1 = require("../../PatchBuilder");
var Decoder = (function (_super) {
    (0, tslib_1.__extends)(Decoder, _super);
    function Decoder() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    Decoder.prototype.decode = function (data) {
        this.reset(data);
        var _a = (0, tslib_1.__read)(this.uint53vuint39(), 2), sessionId = _a[0], time = _a[1];
        var id = (this.patchId = new clock_1.LogicalTimestamp(sessionId, time));
        var clock = new clock_1.LogicalVectorClock(id.getSessionId(), id.time);
        this.builder = new PatchBuilder_1.PatchBuilder(clock);
        this.decodeOperations();
        return this.builder.patch;
    };
    Decoder.prototype.decodeId = function () {
        var patchId = this.patchId;
        var _a = (0, tslib_1.__read)(this.b1vuint56(), 2), flag = _a[0], value = _a[1];
        if (flag) {
            return new clock_1.LogicalTimestamp(patchId.getSessionId(), patchId.time + value);
        }
        else {
            var time = this.vuint57();
            return new clock_1.LogicalTimestamp(value, time);
        }
    };
    Decoder.prototype.decodeOperations = function () {
        while (this.x < this.uint8.length)
            this.decodeOperation();
    };
    Decoder.prototype.decodeOperation = function () {
        var opcode = this.u8();
        switch (opcode) {
            case 0: {
                this.builder.obj();
                return;
            }
            case 1: {
                this.builder.arr();
                return;
            }
            case 2: {
                this.builder.str();
                return;
            }
            case 14: {
                this.builder.val(this.val());
                return;
            }
            case 3: {
                this.builder.num();
                return;
            }
            case 4: {
                this.builder.root(this.decodeId());
                return;
            }
            case 5: {
                var object = this.decodeId();
                var fields = this.vuint57();
                var tuples = [];
                for (var i = 0; i < fields; i++) {
                    var value = this.decodeId();
                    var key = this.val();
                    if (typeof key !== 'string')
                        continue;
                    tuples.push([key, value]);
                }
                this.builder.setKeys(object, tuples);
                return;
            }
            case 15: {
                var after = this.decodeId();
                var value = this.val();
                this.builder.setVal(after, value);
                return;
            }
            case 6: {
                var after = this.decodeId();
                var value = this.val();
                if (typeof value !== 'number')
                    return;
                this.builder.setNum(after, value);
                return;
            }
            case 7: {
                var obj = this.decodeId();
                var after = this.decodeId();
                var str = this.val();
                if (typeof str !== 'string')
                    return;
                this.builder.insStr(obj, after, str);
                return;
            }
            case 8: {
                var arr = this.decodeId();
                var after = this.decodeId();
                var length_1 = this.vuint57();
                var elements = [];
                for (var i = 0; i < length_1; i++)
                    elements.push(this.decodeId());
                this.builder.insArr(arr, after, elements);
                return;
            }
            case 10: {
                var obj = this.decodeId();
                var after = this.decodeId();
                var length_2 = this.vuint57();
                this.builder.del(obj, after, length_2);
                return;
            }
            case 9: {
                var obj = this.decodeId();
                var after = this.decodeId();
                this.builder.del(obj, after, 1);
                return;
            }
            case 11: {
                this.builder.noop(1);
                return;
            }
            case 12: {
                this.builder.noop(this.vuint57());
                return;
            }
            case 16: {
                this.builder.bin();
                return;
            }
            case 17: {
                var obj = this.decodeId();
                var after = this.decodeId();
                var length_3 = this.vuint57();
                var data = this.bin(length_3);
                this.builder.insBin(obj, after, data);
                return;
            }
            default: {
                throw new Error('UNKNOWN_OP');
            }
        }
    };
    return Decoder;
}(CrdtDecoder_1.CrdtDecoder));
exports.Decoder = Decoder;
