"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.AbstractEncoder = void 0;
var tslib_1 = require("tslib");
var ArrayType_1 = require("../../types/rga-array/ArrayType");
var BinaryType_1 = require("../../types/rga-binary/BinaryType");
var ConstantType_1 = require("../../types/const/ConstantType");
var CrdtEncoder_1 = require("../../../json-crdt-patch/util/binary/CrdtEncoder");
var ObjectType_1 = require("../../types/lww-object/ObjectType");
var StringType_1 = require("../../types/rga-string/StringType");
var utf8_1 = require("../../../util/utf8");
var ValueType_1 = require("../../types/lww-value/ValueType");
var AbstractEncoder = (function (_super) {
    (0, tslib_1.__extends)(AbstractEncoder, _super);
    function AbstractEncoder() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    AbstractEncoder.prototype.encodeRoot = function (root) {
        this.ts(root.id);
        if (root.node)
            this.encodeNode(root.node);
    };
    AbstractEncoder.prototype.encodeNode = function (node) {
        if (node instanceof ObjectType_1.ObjectType)
            return this.encodeObj(node);
        else if (node instanceof ArrayType_1.ArrayType)
            return this.encodeArr(node);
        else if (node instanceof StringType_1.StringType)
            return this.encodeStr(node);
        else if (node instanceof ValueType_1.ValueType)
            return this.encodeVal(node);
        else if (node instanceof ConstantType_1.ConstantType)
            return this.encodeConst(node);
        else if (node instanceof BinaryType_1.BinaryType)
            return this.encodeBin(node);
        throw new Error('UNKNOWN_NODE');
    };
    AbstractEncoder.prototype.encodeObj = function (obj) {
        var e_1, _a;
        var length = obj.latest.size;
        this.encodeObjectHeader(length);
        this.ts(obj.id);
        try {
            for (var _b = (0, tslib_1.__values)(obj.latest.entries()), _c = _b.next(); !_c.done; _c = _b.next()) {
                var _d = (0, tslib_1.__read)(_c.value, 2), key = _d[0], chunk = _d[1];
                this.encodeObjChunk(key, chunk);
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
            }
            finally { if (e_1) throw e_1.error; }
        }
    };
    AbstractEncoder.prototype.encodeObjChunk = function (key, chunk) {
        this.ts(chunk.id);
        var length = (0, utf8_1.utf8Count)(key);
        this.vuint57(length);
        this.encodeUtf8(key, length);
        this.encodeNode(chunk.node);
    };
    AbstractEncoder.prototype.encodeArr = function (obj) {
        var e_2, _a;
        var length = obj.size();
        this.encodeArrayHeader(length);
        this.ts(obj.id);
        try {
            for (var _b = (0, tslib_1.__values)(obj.chunks()), _c = _b.next(); !_c.done; _c = _b.next()) {
                var chunk = _c.value;
                this.encodeArrChunk(chunk);
            }
        }
        catch (e_2_1) { e_2 = { error: e_2_1 }; }
        finally {
            try {
                if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
            }
            finally { if (e_2) throw e_2.error; }
        }
    };
    AbstractEncoder.prototype.encodeArrChunk = function (chunk) {
        if (chunk.deleted) {
            this.b1vuint56(true, chunk.deleted);
            this.ts(chunk.id);
        }
        else {
            var nodes = chunk.nodes;
            var length_1 = nodes.length;
            this.b1vuint56(false, length_1);
            this.ts(chunk.id);
            for (var i = 0; i < length_1; i++)
                this.encodeNode(nodes[i]);
        }
    };
    AbstractEncoder.prototype.encodeStr = function (obj) {
        var e_3, _a;
        var length = obj.size();
        this.encodeStringHeader(length);
        this.ts(obj.id);
        try {
            for (var _b = (0, tslib_1.__values)(obj.chunks()), _c = _b.next(); !_c.done; _c = _b.next()) {
                var chunk = _c.value;
                this.encodeStrChunk(chunk);
            }
        }
        catch (e_3_1) { e_3 = { error: e_3_1 }; }
        finally {
            try {
                if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
            }
            finally { if (e_3) throw e_3.error; }
        }
    };
    AbstractEncoder.prototype.encodeStrChunk = function (chunk) {
        if (chunk.deleted) {
            this.b1vuint56(true, chunk.deleted);
            this.ts(chunk.id);
        }
        else {
            var text = chunk.str;
            var length_2 = (0, utf8_1.utf8Count)(text);
            this.b1vuint56(false, length_2);
            this.ts(chunk.id);
            this.encodeUtf8(text, length_2);
        }
    };
    AbstractEncoder.prototype.encodeBin = function (obj) {
        var e_4, _a;
        var length = obj.size();
        this.encodeBinaryHeader(length);
        this.ts(obj.id);
        try {
            for (var _b = (0, tslib_1.__values)(obj.chunks()), _c = _b.next(); !_c.done; _c = _b.next()) {
                var chunk = _c.value;
                this.encodeBinChunk(chunk);
            }
        }
        catch (e_4_1) { e_4 = { error: e_4_1 }; }
        finally {
            try {
                if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
            }
            finally { if (e_4) throw e_4.error; }
        }
    };
    AbstractEncoder.prototype.encodeBinChunk = function (chunk) {
        if (chunk.deleted) {
            this.b1vuint56(true, chunk.deleted);
            this.ts(chunk.id);
        }
        else {
            var data = chunk.buf;
            var length_3 = data.byteLength;
            this.b1vuint56(false, length_3);
            this.ts(chunk.id);
            this.buf(data, length_3);
        }
    };
    AbstractEncoder.prototype.encodeConst = function (obj) {
        var value = obj.value;
        switch (value) {
            case null:
            case false:
            case true:
                this.encodeAny(value);
                break;
            case undefined:
                this.u8(0xc1);
                break;
            default: {
                if (typeof value === 'number')
                    this.encodeNumber(value);
                else {
                    this.u8(0xd4);
                    this.encodeAny(value);
                }
            }
        }
    };
    AbstractEncoder.prototype.encodeVal = function (obj) {
        this.u8(0xd5);
        this.ts(obj.id);
        this.ts(obj.writeId);
        this.encodeAny(obj.value);
    };
    AbstractEncoder.prototype.encodeUtf8 = function (str, byteLength) {
        this.ensureCapacity(byteLength);
        var strLength = str.length;
        var uint8 = this.uint8;
        var offset = this.offset;
        var pos = 0;
        while (pos < strLength) {
            var value = str.charCodeAt(pos++);
            if ((value & 0xffffff80) === 0) {
                uint8[offset++] = value;
                continue;
            }
            else if ((value & 0xfffff800) === 0) {
                uint8[offset++] = ((value >> 6) & 0x1f) | 0xc0;
            }
            else {
                if (value >= 0xd800 && value <= 0xdbff) {
                    if (pos < strLength) {
                        var extra = str.charCodeAt(pos);
                        if ((extra & 0xfc00) === 0xdc00) {
                            pos++;
                            value = ((value & 0x3ff) << 10) + (extra & 0x3ff) + 0x10000;
                        }
                    }
                }
                if ((value & 0xffff0000) === 0) {
                    uint8[offset++] = ((value >> 12) & 0x0f) | 0xe0;
                    uint8[offset++] = ((value >> 6) & 0x3f) | 0x80;
                }
                else {
                    uint8[offset++] = ((value >> 18) & 0x07) | 0xf0;
                    uint8[offset++] = ((value >> 12) & 0x3f) | 0x80;
                    uint8[offset++] = ((value >> 6) & 0x3f) | 0x80;
                }
            }
            uint8[offset++] = (value & 0x3f) | 0x80;
        }
        this.offset = offset;
    };
    return AbstractEncoder;
}(CrdtEncoder_1.CrdtEncoder));
exports.AbstractEncoder = AbstractEncoder;
